{% extends "!layout.html" %}
{% set script_files = ['_static/gwpy_https.js'] + script_files + ['_static/copybutton.js']%}
{% set bootswatch_css_custom = ['//gwpy.github.io/css/gwpy.css', '_static/gwpy-sphinx.css']%}
