# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2015)

# package metadata
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = '0.1b4'
__date__ = '2015-11-18 22:21:18 +0000'

# package version
version = '0.1b4'
major = 0
minor = 1
micro = None
debug = True
release = False

# repository version information
git_hash = '25914f2e2f31ba62ad3d2ea246697925ab4d5346'
git_tag = 'v0.1b4'
git_branch = 'master'
git_author = "Duncan Macleod <duncanmmacleod@gmail.com>"
git_committer = "Duncan Macleod <duncanmmacleod@gmail.com>"
git_is_dirty = False